/*
 * JIO extension for resource replication.
 * Copyright (C) 2013  Nexedi SA
 *
 *   This library is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*jslint indent: 2, maxlen: 80, nomen: true */
/*global define, module, require, jIO, RSVP, JSON */

(function (root, dependencies, factory) {
  "use strict";
  if (typeof define === 'function' && define.amd) {
    return define(dependencies, function () {
      return factory(require);
    });
  }
  if (typeof require === 'function') {
    module.exports = factory(require);
    return;
  }
  root.replicate_storage = factory(function (name) {
    return {
      "jio": jIO,
      "rsvp": RSVP
    }[name];
  });
}(this, ['jio', 'rsvp'], function (require) {
  "use strict";

  var Promise = require('rsvp').Promise,
    all = require('rsvp').all,
    dictUpdate = require('jio').util.dictUpdate,
    addStorageFunction = require('jio').addStorage;

  /**
   * Make a clone of an object WITHOUT _xxx (except _id). This is what is
   * compared when compared when checking answers from multiple storages
   * @method cleanClone
   * @param {Object} obj to clone
   * @return {Object} clones object
   **/
  function cleanClone (obj) {
    var prop, r = {};

    for (prop in obj) {
      if (obj.hasOwnProperty(prop)) {
        if (prop.charCodeAt(0) !== 95 || prop === "_id") {
          r[prop] = obj[prop];
        }
      }
    }
    return r;
  };

  /**
   * Generate a checksum from a passed string to compare the response returned
   * by multiple storages for the same requests.
   * @method checksum
   * @param {Object} obj to generate a checksum for
   * @return {Integer} checksum
   **/
  function checksum (obj) {
    var r, i, str = JSON.stringify(obj);

    for(r = 0, i = 0; i < str.length; i += 1) {
      r = (r << 5) - r + str.charCodeAt(i);
      r &= r;
    }
    return r;
  };

  /**
   * Executes a sequence of *then* callbacks. It acts like
   * `smth().then(callback).then(callback)...`. The first callback is called
   * with no parameter.
   *
   * Elements of `then_list` array can be a function or an array contaning at
   * most three *then* callbacks: *onFulfilled*, *onRejected*, *onNotified*.
   *
   * When `cancel()` is executed, each then promises are cancelled at the same
   * time.
   *
   *     sequence(then_list): Promise
   *
   * @param  {Array} then_list An array of *then* callbacks
   * @return {Promise} A new promise
   */
  function sequence(then_list) {
    var promise_list = [];
    return new Promise(function (resolve, reject, notify) {
      var i, length = then_list.length;
      promise_list[0] = new Promise(function (resolve) {
        resolve();
      });
      for (i = 0; i < length; i += 1) {
        if (Array.isArray(then_list[i])) {
          promise_list[i + 1] = promise_list[i].
            then(then_list[i][0], then_list[i][1], then_list[i][2]);
        } else {
          promise_list[i + 1] = promise_list[i].then(then_list[i]);
        }
      }
      promise_list[i].then(resolve, reject, notify);
    }, function () {
      var i, length = promise_list.length;
      for (i = 0; i < length; i += 1) {
        promise_list[i].cancel();
      }
    });
  }

  function success(promise) {
    return new Promise(function (resolve, reject, notify) {
      /*jslint unparam: true*/
      promise.then(resolve, resolve, notify);
    }, function () {
      promise.cancel();
    });
  }

  /**
   * Responds with the last resolved promise answer recieved. If all promises
   * are rejected, it returns the latest rejected promise answer
   * received. Promises are cancelled only by calling
   * `last(promise_list).cancel()`.
   *
   *     last(promise_list): Promise
   *
   * @param  {Array} promise_list An array of promises
   * @param  {Integer} master The "promise" with id authority
   * @param  {Object} temp Any temporary data created by master
   * @return {Promise} A new promise
   */
  // NOTE: added master parameter for master-slave storage
  function last(promise_list, master, temp) {
    var length = promise_list.length;
    promise_list = promise_list.slice();

    return new Promise(function (resolve, reject, notify) {
      var index, last_answer, count = 0, error_count = 0;

      function resolver() {
        return function (answer) {
          count += 1;
          if (count === length) {
            if (temp) {
              answer._temp = temp;
            }
            return resolve(answer);
          }
          last_answer = answer;
        };
      }
      function rejecter() {
        return function (answer) {
          error_count += 1;
          if (error_count === length) {
            return reject(answer);
          }
          count += 1;
          if (count === length) {
            if (temp) {
              answer._temp = temp;
            }
            return resolve(last_answer);
          }
        };
      }
      function notifier(index) {
        return function (notification) {
          notify({
            "index": index,
            "value": notification
          });
        };
      }

      if (typeof master === "number") {
        length = 1;
        promise_list[master].then(resolver(), rejecter(), notifier(master));
      } else {
        for (index = 0; index < length; index += 1) {
          promise_list[index].then(resolver(), rejecter(), notifier(index));
        }
      }
    }, function () {
      var index;
      for (index = 0; index < length; index += 1) {
        promise_list[index].cancel();
      }
    });
  }

  /**
   * Test responses of a GET request for coherence (will set _conflict)
   * and missing (will set _missing) on the returned document
   * @method checkAndReturn
   * @param {Object} promise_list List of responses received
   * @return {Promise} A new promise
   **/
  function checkAndReturn(promise_list) {
    var length = promise_list.length;

    promise_list = promise_list.slice();
    return new Promise(function (resolve, reject, notify) {
      var index, trace, conflict_list, count, error_count, tag, doc;

      count = 0;
      error_count = 0;
      conflict_list = [];
      tag = function (reply) {
        if (conflict_list.length) {
          reply._conflict_list = conflict_list;
        }
        if (error_count) {
          reply._missing = true;
        }
        return reply;
      };

      function resolver(answer) {
        if (trace === undefined) {
          doc = answer;
          trace = checksum(cleanClone(answer.data));
        } else {
          if (trace !== checksum(cleanClone(answer.data))) {
            conflict_list.push(cleanClone(answer.data));
          }
        }
        count += 1;
        if (count === length) {
          return resolve(tag(answer));
        }
      }
      function rejecter(answer) {
        error_count += 1;
        if (error_count === length) {
          return reject(answer);
        }
        count += 1;
        if (count === length) {
          return resolve(tag(doc));
        }
      }
      function notifier(index) {
        return function (notification) {
          notify({
            "index": index,
            "value": notification
          });
        };
      }
      for (index = 0; index < length; index += 1) {
        promise_list[index].then(resolver, rejecter, notifier(index));
      }
    }, function () {
      var index;
      for (index = 0; index < length; index += 1) {
        promise_list[index].cancel();
      }
    });
  }


  function ReplicateStorage(spec) {
    if (!Array.isArray(spec.storage_list)) {
      throw new TypeError("ReplicateStorage(): " +
                          "storage_list is not of type array");
    }
    this._storage_list = spec.storage_list;
    this._master = spec.master;
  };


  ReplicateStorage.prototype.post = function (command, metadata, option) {
    var index, subindex,
      storages = this._storage_list,
      promise_list = [],
      priority_list = [],
      length = storages.length,
      master = this._master;

    // first run priority storage
    if (master !== undefined) {
      priority_list[0] = success(
        command.storage(storages[master]).post(metadata, option)
      );
    } else {
      for (index = 0; index < length; index += 1) {
        promise_list[index] = success(
          command.storage(storages[index]).post(metadata, option)
        );
      }
    }

    // first master, then slave
    sequence([
      function () {
        if (master !== undefined) {
          return last(priority_list, master);
        }
        return {};
      },
      function (response) {
        var method, subindex = 0;
        if (response.id) {
          metadata._id = response.id;
          method = "put";
        } else {
          // TODO: no-access master 404/offline handling
        }

        for (index = 0; index < length; index += 1) {
          if (index !== master) {
            promise_list[subindex] = success(
              command.storage(storages[index])
                  [method || "post"](metadata, option)
            );
            subindex += 1;
          }
        }
        return last(promise_list);
      },
      [command.success, command.error]
    ]);
  };

  ReplicateStorage.prototype.put = function (command, metadata, option) {
    var index, subindex,
      storages = this._storage_list,
      promise_list = [],
      priority_list = [],
      length = storages.length,
      master = this._master;

    // setup master storage or all
    if (master !== undefined) {
      priority_list[0] = success(
        command.storage(storages[master]).put(metadata, option)
      );
    } else {
      for (index = 0; index < length; index += 1) {
        promise_list[index] = success(
          command.storage(storages[index]).put(metadata, option)
        );
      }
    }

    // run master, then slave
    sequence([
      function () {
        return last(priority_list, master);
      },
      function (response) {
        var subindex;

        switch (true) {
          // NOTE: this means if the master is 503, nothing will be saved on
          // the slave? Is this what I would want? YES
          // NOTE: leaving this will allow test pages to be saved, but
          // will break at another place where 500 should not pass
          case response.status >= 500:
//             command.error(response);
//             break;
          // HACK: why does 204 return status code 0 on promises?
          case response.status === 0:
          default:
            for (index = 0, subindex = 0; index < length; index += 1) {
              if (index !== master) {
                promise_list[subindex] = success(
                  command.storage(storages[index]).put(metadata, option)
                );
                subindex += 1;
              }
            }
            return last(promise_list, undefined, response._temp);
        }
      },
      [command.success, command.error]
    ]);
  };

  ReplicateStorage.prototype.putAttachment = function (command, param, option) {
    var promise_list = [], index, length = this._storage_list.length;
    for (index = 0; index < length; index += 1) {
      promise_list[index] = success(
        command.storage(this._storage_list[index]).putAttachment(param, option)
      );
    }
    sequence([function () {
      return last(promise_list);
    }, [command.success, command.error]]);
  };

  ReplicateStorage.prototype.remove = function (command, param, option) {
    var promise_list = [], index, length = this._storage_list.length;
    for (index = 0; index < length; index += 1) {
      promise_list[index] = success(
        command.storage(this._storage_list[index]).remove(param, option)
      );
    }
    sequence([function () {
      return last(promise_list);
    }, [command.success, command.error]]);
  };

  ReplicateStorage.prototype.removeAttachment = function (
    command,
    param,
    option
  ) {
    var promise_list = [], index, length = this._storage_list.length;
    for (index = 0; index < length; index += 1) {
      promise_list[index] = success(
        command.storage(this._storage_list[index]).
          removeAttachment(param, option)
      );
    }
    sequence([function () {
      return last(promise_list);
    }, [command.success, command.error]]);
  };

  ReplicateStorage.prototype.get = function (command, param, option) {
    var promise_list = [], index, length = this._storage_list.length;
    for (index = 0; index < length; index += 1) {
      promise_list[index] =
        command.storage(this._storage_list[index]).get(param, option);
    }
    sequence([function () {
      return checkAndReturn(promise_list);
    }, [command.success, command.error]]);
  };

  ReplicateStorage.prototype.getAttachment = function (command, param, option) {
    var promise_list = [], index, length = this._storage_list.length;
    for (index = 0; index < length; index += 1) {
      promise_list[index] = success(
        command.storage(this._storage_list[index]).getAttachment(param, option)
      );
    }
    sequence([function () {
      return last(promise_list);
    }, [command.success, command.error]]);
  };

  ReplicateStorage.prototype.allDocs = function (command, param, option) {
    /*jslint unparam: true */
    var promise_list = [], index, length = this._storage_list.length;
    for (index = 0; index < length; index += 1) {
      promise_list[index] =
        success(command.storage(this._storage_list[index]).allDocs(option, param));
    }

    sequence([function () {
      return all(promise_list);
    }, function (answers) {
      var i, j, k, l, record, reply, count, total, base, test, must_add,
        empty_reply, rows_len, rows, len = answers.length;

      for (i = 0; i < len; i += 1) {
        reply = answers[i];
        if (reply.result === "success") {
          total = reply.data.total_rows;
          if (total === 0) {
            empty_reply = true;
          }

          // loop records returned in response
          for (j = 0; j < total; j += 1) {
            record = reply.data.rows[j];
            must_add = undefined;

            if (rows === undefined) {
              rows = [record];
            } else {
              count = rows.length;

              // loop records already in rows object
              for (k = 0; k < count; k += 1) {
                base = rows[k];

                // record exists, test for coherence
                if (base.id === record.id) {
                  must_add = undefined;
                  if (checksum(cleanClone(base)) !== checksum(record)) {
                    if (!base._conflict_list) {
                      base._conflict_list = [base];
                    }
                    base._conflict_list.push(record);
                  }
                // record does not exists, must add
                } else {
                  must_add = true;
                }
              }
              // add missing record
              if (must_add) {
                // if on 2+ iteration, this means a 404 on previous storages
                if (i > 0) {
                  record._missing = true;
                }
                rows.push(record);
              }
            }
          }
        }
      }

      // if a storage returns rows = 0, make sure all records are labelled
      rows_len = (rows || []).length;
      if (empty_reply && rows_len > 0) {
        for (l = 0; l < rows_len; l += 1) {
          rows[l]._missing = true;
        }
      }
      return {"data": {"total_rows": (rows || []).length, "rows": rows || []}};
    }, [command.success, command.error]]);
  };

  ReplicateStorage.prototype.check = function (command, param, option) {
    return command.error(501);
  };

  ReplicateStorage.prototype.repair = function (command, param) {
    var i, promise_list, length, index, that;

    promise_list = [];
    length = this._storage_list.length;

    // NOTE: we can't pass full document, because of select_list...
    for (i = 0; i < length; i += 1) {
      promise_list[i] = command.storage(this._storage_list[i]).put(param);
    }

    sequence([function () {
      return all(promise_list);
    }, function (answers) {
      var i;
      for (i = 0; i < answers.length; i += 1) {
        if (answers[i].result !== "success") {
          return command.error(409);
        }
      }
      command.success();
    }]);
  };

  addStorageFunction('replicate', ReplicateStorage);

}));
